/* 
   Copyright (C) 1999 Free Software Foundation, Inc.

This file is part of Nilo.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */

/* This is generated by running configure */
#include "niloconfig.h"

/* Define the basic data types */
#include <sys/types.h>

/* Define the PXE types */
#include "pxe_types.h"

/* Define the tracing macros */
#include "trace.h"

/* Define the PXE TFTP data structures */
#include "pxe_tftp.h"

/* Define the PXE UDP data structures */
#include "pxe_udp.h"

#include "pxe.h"
#include "pxenv.h"

#include "nilo.h"

#include "tftp.h"

extern pxe_t *pxe_data_struct;
extern pxenv_t *pxenv_data_struct;

static tftp_t tftp_pkt;

/* pxenv_tftp_open() - open the TFTP connection so we can read and write.
 * 
 * parameters:
 *     tftpopen.server_ip - the IP address of the TFTP server
 *                          in Big Endian Format.
 *     tftpopen.gw_ip - the IP address of gateway in Big Endian
 *                      format.
 *     tftpopen.filename - a NULL terminated string
 *                         representing the boot file name.
 *     tftpopen.tftpport - the port the TFTP server is
 *                         listening on. This must be in Big Endian
 *                         format.
 *     tftpopen.packetsize - the requested size of the TFTP
 *                           packets. This must be at least 512 bytes.
 * returns:
 *     tftpopen.status - the exit status of this function. This is set to
 *                       one of the PXENV_STATUS_xxx constants.
 *     tftpopen.packetsize - the negotiated size of the TFTP packets. Must
 *                           be at least 512 bytes.
 */
pxenv_ret_code_t
pxenv_tftp_open (pxenv_tftp_open_t *tftpopen)
{
    pxenv_udp_read_t udpread;
    u_int32_t timeout;
    int snaplen;

    NILO_TRACE_FUNCTION;

    client_info->tftp_blksize = TFTP_DEFAULTSIZE_PACKET;
    client_info->tftp_client_port = tftpopen->tftpport;
    client_info->tftp_blockno = 1; /* TFTP sessions start with a DATA block of 1 */
    switch (client_info->client_state) {
        /* If we've finished the other two major states, make a
           request for the file. */
    case DHCP_DONE:
    case ARPING:		/* FIXME: We shouldn't need this here */
    case ARP_DONE:
        client_info->client_state = TFTP_REQUEST;
        NILO_TRACE_MSG("Requesting file %s via tftp.\r\n", tftpopen->filename);
        snaplen = fill_tftp_request(&tftp_pkt, TFTP_RRQ, tftpopen->filename,
                                    0);
	/* We don't use the packetsize here, and we should. This will
	   break when MTFTPD is used. */
	//	client_info->tftp_client_port++;
        send_tftp_request(client_info, &tftp_pkt, snaplen, 1, 100);
        client_info->client_state = TFTP_LOADING;
        break;
    }

    NILO_TRACE_FUNCTION_RETURN;
    return (PXENV_EXIT_SUCCESS);
}

/* pxenv_tftp_close() - close the open TFTP connection.
 *
 * parameters:
 *     none
 * returns:
 *     tftpclose.status - the exit status of this function. This is set
 *                        to one of the PXENV_STATUS_xxx constants.
 */
pxenv_ret_code_t
pxenv_tftp_close (pxenv_tftp_close_t *tftpclose)
{
    NILO_TRACE_FUNCTION;

    NILO_TRACE_FUNCTION_RETURN;
    return (PXENV_EXIT_SUCCESS);
}

/* pxenv_tftp_read() - read one packet from an open TFTP connection.
 *
 * parameters:
 *     tftpread.buffersize - the size of the receive buffer in bytes. 
 *     tftpread.buffer - the buffer address.
 *     tftpread.packetnum - the packet sequence number.
 * returns:
 *     tftpread.status - the exit status of this function. This is set to
 *                       one of the PXENV_STATUS_xxx constants.
 *     tftpread.packetnum - the packet sequence number.
 *     tftpread.buffersize - the size of the packet received in
 *                           the buffer.
 */
pxenv_ret_code_t
pxenv_tftp_read (pxenv_tftp_read_t *tftpread)
{
    pxenv_udp_read_t udpread;
    u_int32_t timeout, retries;
    tftp_t tftp_pkt, *tftp_ptr;

    NILO_TRACE_FUNCTION;

    switch (client_info->client_state) {
      /* If we've finished the other two major states, make a
	 request for the file. */
    case TFTP_DONE:
        NILO_TRACE_FUNCTION_RETURN;
        return PXENV_EXIT_SUCCESS;
    case TFTP_LOADING:      
      retries = 5;
      while (retries--) {
	udpread.dest_ip.num = client_info->yiaddr.num;
	udpread.d_port = client_info->tftp_client_port;
	udpread.buffer = tftpread->buffer;
	udpread.buffer_size = sizeof(tftpread->buffer);
	if (call_pxe_api (PXENV_UDP_READ, &udpread) == PXENV_EXIT_FAILURE) {
	  NILO_TRACE_MSG ("Couldn't read a UDP packet.\r\n");
          tftpread->buffersize = 0;
	  timeout = currticks() + 32;
	  while (timeout > currticks()) { }
	  continue;
	} else {
	  tftpread->buffersize = udpread.buffer_size;
          tftp_ptr = (tftp_t *)udpread.buffer;
	  tftp_decode_packet ((tftp_t *)udpread.buffer, udpread.buffer_size);
          /* Check to make sure we're not overrunning the server */
          if (ntohs(tftp_ptr->u.data.block) > client_info->tftp_blockno +2) {
              NILO_TRACE_MSG ("Packet numbers too far out of sync.\r\n");
              client_info->client_state = TFTP_DONE;
              NILO_TRACE_FUNCTION_RETURN;
              return PXENV_EXIT_FAILURE;
          }
	  break;
	}
        }
      if (retries <= 0)
          tftpread->buffersize = 0;
      //      client_info->client_state = TFTP_DONE; /* FIXME: HACK ALERT!! */
    case TFTP_ACKING:
      NILO_TRACE_MSG("Ack of tftp block 0x%02x\r\n",
		     client_info->tftp_blockno);
      fill_tftp_request(&tftp_pkt, TFTP_ACK, 
                        client_info->file,
                        tftpread->packetnum);
      send_tftp_request(client_info, &tftp_pkt, sizeof(tftp_t), 1, 100);
      client_info->client_state = TFTP_LOADING;
      break;
    }
    NILO_TRACE_FUNCTION_RETURN;
    return (PXENV_EXIT_SUCCESS);
}

/* pxenv_tftp_read_file() - open the TFTP connection, load the entire
 *      file, and close the connection.
 *
 * parameters:
 *     tftpfile.filename - a NULL terminated string represnting the boot
 *                         file name.
 *     tftpfile.buffersize - the size of the receive buffer in bytes. 
 *     tftpfile.buffer - the buffer address.
 *     tftpfile.server_ip - the IP address of the TFTP server in Big Endian
 *                          format.
 *     tftpfile.gw_ip - the IP address of gateway in Big Endian
 *                      format.
 *     tftpfile.mcast_ip - the multicast IP address in Big Endian format.
 *     tftpfile.tftp_clnt_port - the port the client listens to for a
 *                               multicast.
 *     tftpfile.tftp_srv_port - the port the server listens to for a
 *                              multicast.
 *     tftpfile.tftp_open_timeout - the timeout value in seconds used for
 *                                  receiving data. If zero, the default of
 *                                  TFTP timeout is used.
 *     tftpfile.tftp_reopen_delay - the maximum wait delay time in seconds
 *                                  between ACK of a fast packet and a new
 *                                  TFTP open.
 * returns:
 *     tftpfile.status - The exit status of this function. This is set to one
 *                       of the PXENV_STATUS_xxx constants.
 *     tftpfile.buffersize - the size of the packet received in the buffer.
 */
pxenv_ret_code_t
pxenv_tftp_read_file (pxenv_tftp_read_file_t *tftpfile)
{
    pxenv_tftp_open_t tftpopen;
    pxenv_tftp_read_t tftpread;
    int snaplen;
    
    NILO_TRACE_FUNCTION;

    /* open the TFTP connection to the server */
    tftpopen.server_ip = tftpfile->server_ip;/* server IP number */
    tftpopen.gw_ip = tftpfile->gw_ip;    /* gateway IP number */
    strcpy(tftpopen.filename, tftpfile->filename); /* boot file name */
    tftpopen.tftpport = tftpfile->tftp_clnt_port; /* TFTP port */
    if (client_info->tftp_blksize)
        tftpopen.packetsize = client_info->tftp_blksize; /* TFTP blocksize */

    if (CALL_PXE_API (PXENV_TFTP_OPEN, &tftpopen) == PXENV_EXIT_FAILURE) {
        panic ("Couldn't open the TFTP connection.\r\n");
        return (PXENV_EXIT_FAILURE);
    }

    /* read the file, one packet at a time */
    while (client_info->client_state != TFTP_DONE) {
	tftpread.buffer = &tftp_pkt;
	tftpread.packetnum = client_info->tftp_blockno++;
        tftpread.buffersize = client_info->tftp_blksize; /* TFTP blocksize */
	if (CALL_PXE_API (PXENV_TFTP_READ, &tftpread) == PXENV_EXIT_FAILURE) {
	  panic ("Couldn't read from the TFTP connection.\r\n");
	  return (PXENV_EXIT_FAILURE);
	}
	if (tftpread.buffersize == 0) {
            NILO_TRACE_PRINTF ("\tRequested a block size of %d, got %d\r\n",
                               tftpfile->buffersize, tftpread.buffersize);
            NILO_TRACE_MSG ("End of file reached.\r\n");
            break;
	}
    }
    //    client_info->tftp_blockno++;
    tftpfile->buffersize = tftpread.buffersize;
    
    NILO_TRACE_FUNCTION_RETURN;
    return (PXENV_EXIT_SUCCESS);
}

/* pxenv_tftp_get_fsize() - get the size of the file to download. This must
 *      not be called when there is an open connection.
 *
 * parameters:
 *     tftpfsize.server_ip - the IP address of the TFTP server in Big Endian
 *                 format.
 *     tftpfsize.gw_ip - The IP address of gateway in Big Endian
 *             format.
 *     tftpfsize.filename - a NULL terminated string represnting the boot
 *             file name.
 *     tftpfsize.tftpport - the port the TFTP server is listening on. This
 *                          must be in Big Endian format.
 * returns:
 *     tftpfsize.status - The exit status of this function. This is set
 *                        to one of the PXENV_STATUS_xxx constants.
 *     tftpfsize.filesize - the size of the boot file in bytes.
 */
pxenv_ret_code_t
pxenv_tftp_get_fsize (pxenv_tftp_get_fsize_t *tftpfsize)
{
    NILO_TRACE_FUNCTION;

    NILO_TRACE_FUNCTION_RETURN;
    return (PXENV_EXIT_SUCCESS);
}


